/**
 * @author Peter Nitsch
 */

TERM.Session = function(fontMapURL, updateStateFunc) {

	var viewer;
	var commands, term_state = 'disconnected';
	var handle_message, handle_oob_message, normal_msg, init_msg, ovm_init,ovm_init_oob, updateState;
	var term_host,term_port;
	var vmName;

	var onUpdateState = updateStateFunc;

	updateState = function(state, statusMsg) {
		var func, cmsg, oldstate = term_state;

		if (state === oldstate) {
			/* Already here, ignore */
			Util.Debug("Already in state '" + state + "', ignoring.");
			return;
		}

		/*
		 * These are disconnected states. A previous connect may asynchronously
		 * cause a connection so make sure we are closed.
		 */
		if (state in {
			'disconnected' : 1,
			'loaded' : 1,
			'connect' : 1,
			'disconnect' : 1,
			'failed' : 1,
			'fatal' : 1
		}) {

			if (TERM.socket != null) {
				Util.Info("Going to CLOSE Websocket");
				TERM.socket.close();
				Util.Info("CLOSED Websocket");
			}
			if (TERM.oob_socket != null) {
				Util.Info("Going to CLOSE Oob Websocket");
				TERM.oob_socket.close();
				Util.Info("CLOSED Oob Websocket");
			}

		}

		if (oldstate === 'fatal') {
			Util.Error("Fatal error, cannot continue");
		}

		if ((state === 'failed') || (state === 'fatal')) {
			func = Util.Error;
		} else {
			func = Util.Warn;
		}

		cmsg = typeof (statusMsg) !== 'undefined' ? (" Msg: " + statusMsg) : "";
		func("New state '" + state + "', was '" + oldstate + "'." + cmsg);

		if ((oldstate === 'failed') && (state === 'disconnected')) {
			// Do disconnect action, but stay in failed state
			term_state = 'failed';
		} else {
			term_state = state;
		}

		switch (state) {
		case 'normal':
			if ((oldstate === 'disconnected') || (oldstate === 'failed')) {
				Util
						.Error("Invalid transition from 'disconnected' or 'failed' to 'normal'");
			}

			break;

		case 'connect':

			/*
			 * connTimer = setTimeout(function () { fail("Connect timeout"); },
			 * conf.connectTimeout * 1000);
			 */

			// connect();
			break;

		case 'disconnect':

			/*
			 * if (! test_mode) { disconnTimer = setTimeout(function () {
			 * fail("Disconnect timeout"); }, conf.disconnectTimeout * 1000); }
			 */

			// print_stats();
			// WebSocket.onclose transitions to 'disconnected'
			break;

		case 'failed':
			if (oldstate === 'disconnected') {
				Util
						.Error("Invalid transition from 'disconnected' to 'failed'");
			}
			if (oldstate === 'normal') {
				Util.Error("Error while connected.");
			}
			if (oldstate === 'init') {
				Util.Error("Error while initializing.");
			}

			// Make sure we transition to disconnected
			setTimeout(function() {
				updateState('disconnected');
			}, 50);

			break;

		case 'ovs_disconnect':

			break;
		default:
			// No state change action to take

		}

		if ((oldstate === 'failed') && (state === 'disconnected')) {
			// Leave the failed message
			// figure out how to do this
			onUpdateState(state, oldstate);
		} else {
			onUpdateState(state, oldstate, statusMsg);
		}
	};

	function initKeyboard() {
		document.addEventListener("keypress", function(event) {
			var char = null;
			if (event.which == null)
				char = event.keyCode; // old IE
			else if (event.which != 0 && event.charCode != 0)
				char = event.which; // All others
			
			if (char === KEYBOARD_ENTER) {
				return; // do not process keyboard enter here, its being done in
				// keydown. (ff doesnt enter in here, but chrome does)
			}
			if (char != null) {
				TERM.socket.send(char);
			}
			event.preventDefault();

		}, false);

		document.addEventListener("keydown", function(e) {
			var key = e.keyCode;
			var handled= false;
			switch (key) {
			case KEYBOARD_ENTER:
				TERM.socket.send(KEYBOARD_ENTER);
				handled = true;
				break;
			case KEYBOARD_LEFT:
				var key=[];
				key.push(ESCAPE);
				if (viewer.cursorKeyAppMode) {
					key.push(LEFT_SQUARE_BRACKET);
				} else {
					key.push(LATIN_CAPITAL_LETTER_O);
				}
				key.push(LATIN_CAPITAL_LETTER_D);
				TERM.socket.send(key);
				handled = true;
				break;

			case KEYBOARD_RIGHT:
				var key=[];
				key.push(ESCAPE);
				if (viewer.cursorKeyAppMode) {
					key.push(LEFT_SQUARE_BRACKET);
				} else {
					key.push(LATIN_CAPITAL_LETTER_O);
				}
				key.push(LATIN_CAPITAL_LETTER_C);
				TERM.socket.send(key);
				

				handled = true;
				break;

			case KEYBOARD_UP:
				var key=[];
				key.push(ESCAPE);
				if (viewer.cursorKeyAppMode) {
					key.push(LEFT_SQUARE_BRACKET);
				} else {
					key.push(LATIN_CAPITAL_LETTER_O);
				}
				key.push(LATIN_CAPITAL_LETTER_A);
				TERM.socket.send(key);
				handled = true;
				break;

			case KEYBOARD_DOWN:
				var key=[];
				key.push(ESCAPE);
				if (viewer.cursorKeyAppMode) {
					key.push(LEFT_SQUARE_BRACKET);
				} else {
					key.push(LATIN_CAPITAL_LETTER_O);
				}
				key.push(LATIN_CAPITAL_LETTER_B);
				
				TERM.socket.send(key);
				handled = true;
				break;
				
			case KEYBOARD_TAB:
				TERM.socket.send(KEYBOARD_TAB);
				handled=true;
				break;

			case KEYBOARD_PAGE_UP:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(LATIN_CAPITAL_LETTER_M);
				handled = true;
				break;

			case KEYBOARD_PAGE_DOWN:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(LATIN_CAPITAL_LETTER_H);
				TERM.socket.send(SEMICOLON);
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_TWO);
				TERM.socket.send(LATIN_CAPITAL_LETTER_J);
				handled = true;
				break;

			case KEYBOARD_HOME:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(LATIN_CAPITAL_LETTER_H);
				handled = true;
				break;
			case BACKSPACE:
				TERM.socket.send(DELETE);
				handled = true;
				break;
			case LATIN_CAPITAL_LETTER_C:
				if (e.ctrlKey) {
					// Ctrl-C pressed
					TERM.socket.send(LM_SLC_IP);
					handled = true;
				}
				break;
			case KEYBOARD_ESC:
				TERM.socket.send(ESCAPE);
				handled = true;
				break;
			case KEYBOARD_F1:
				//^[[11~				
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F2:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(DIGIT_TWO);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F3:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(DIGIT_THREE);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F4:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(DIGIT_FOUR);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F5:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(DIGIT_FIVE);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F6:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(DIGIT_SEVEN); //Not a mistake
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F7:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(DIGIT_EIGHT);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F8:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(DIGIT_NINE);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F9:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_TWO);
				TERM.socket.send(DIGIT_ZERO);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F10:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_TWO);
				TERM.socket.send(DIGIT_ONE);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F11:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_TWO);
				TERM.socket.send(DIGIT_THREE); //not a mistake
				TERM.socket.send(TILDE);
				handled = true;
				break;
			case KEYBOARD_F12:
				TERM.socket.send(ESCAPE);
				TERM.socket.send(LEFT_SQUARE_BRACKET);
				TERM.socket.send(DIGIT_TWO);
				TERM.socket.send(DIGIT_FOUR);
				TERM.socket.send(TILDE);
				handled = true;
				break;
			default:
				if (key >= 32 && key <= 127) {
					if (e.ctrlKey) {
						switch (key) {
						    // For historic reasons, some control characters are treated specially
					        case /* 3 */ 51: key  =  27; break;
					        case /* 4 */ 52: key  =  28; break;
					        case /* 5 */ 53: key  =  29; break;
					        case /* 6 */ 54: key  =  30; break;
					        case /* 7 */ 55: key  =  31; break;
					        case /* 8 */ 56: key  = 127; break;
					        case /* ? */ 63: key  = 127; break;
					        default:         key &=  31; break;
						}
						TERM.socket.send(key);
						handled = true;
					}
				}
				break;
			}
			if (handled) {
				if (e.preventDefault !== undefined )
					e.preventDefault();
				if (e.stopPropagation !== undefined )
					e.stopPropagation();
			}
		}, false);
	}
	;

	updateState('loaded', "Loaded Ovm Terminal");
	var fontmap = new Image();
	fontmap.onload = function() {
		viewer = new TERM.AnsiViewer(this);
		commands = new TERM.Telnet(viewer);

		if (!("WebSocket" in window)) {
			alert("Sorry, the build of your browser does not support HTML5 WebSockets.");
			return;
		}

		initKeyboard();

	};
	fontmap.src = fontMapURL;

	ovm_init = function(vmachineid, sessionid) {

		Util.Debug(">>> OVM initialisation");
		var connectString = "00360002" + vmachineid;

		var loginData = "sessionid=" + sessionid + "\nconnectString="
				+ connectString;

		var loginDataLength = [];
		loginDataLength.push32(loginData.length);
		TERM.socket.send(loginDataLength);
		TERM.socket.send_string(loginData);

		updateState('OvmInit', "Starting OVM handshaking");

		Util.Debug("<<< OVM initialisation");
	};

	this.connect = function(host, port, vmachineid, sessionid, name) {
		Util.Debug(">>> Session connect");
		term_host = host;
		term_port= port;
		vmName = name;
		if (TERM.socket != undefined && TERM.socket.readyState == 1) {
			viewer.displayCleared();
			viewer.reposition(0, 0);
			Util.Debug(">>> socket readystate sending Init message");

		} else if (TERM.socket == undefined) {
			TERM.socket = new Websock();
			TERM.socket.on('message', handle_message);
			TERM.socket.on('open', function() {

				ovm_init(vmachineid, sessionid);

			});
			TERM.socket.on('close', function(e) {
				Util.Warn("WebSocket on-close event");
				var msg = "";
				if (e.code) {
					msg = " (code: " + e.code;
					if (e.reason) {
						msg += ", reason: " + e.reason;
					}
					msg += ")";
				}

				if (term_state === 'disconnect') {
		            updateState('disconnected', 'VNC disconnected' + msg);
		        } else if (term_state in {'failed':1, 'disconnected':1}) {
		            Util.Error("Received onclose while disconnected" + msg);
		        } else  {
		            fail('Server disconnected' + msg);
		        }
				
				Util.Error("Received onclose : " + msg);
				

			});
			TERM.socket.on('error', function(e) {
				Util.Warn("WebSocket on-error event : " + e.data);
			});

			TERM.oob_socket = new Websock();
			TERM.oob_socket.on('message', handle_oob_message);
			TERM.oob_socket.on('open', function() {
				ovm_init_oob(vmachineid, sessionid);
			});

			var uri;

			uri = "wss://" + host + ":" + port 
					+ "/ovm/core/console/proxy";

			Util.Info("connecting to " + uri);
			updateState('connect', 'Connecting to OVM');

			try {
				TERM.socket.open(uri, [ 'binary', 'base64' ]);
			} catch (Error) {
				Util.Error(Error.message);
		    	fail("Error connecting WebSocket to Oracle VM Manager : " + Error.message);
			}

			Util.Info("<< Session.connect");
		}

	};
	
	function connect_oob() {
		Util.Debug(">> OOB.connect");
		var uri;
		
		uri = "wss://" + term_host + ":" + term_port + "/ovm/core/console/proxy-oob";
		
		Util.Info("connecting to " + uri);
		// TODO: make protocols a configurable
		TERM.oob_socket.open(uri, ['binary', 'base64']);
		
		// Send OVM initialisation
		Util.Debug("<< OOB.connect");
	}
	
	ovm_init_oob = function(vmachineid,sessionid) {
		Util.Info("Initializing OVM Oob Connection");
		
		var connectString = "00360001" + vmachineid;
		
	    var loginData = "sessionid=" + sessionid +"\nconnectString="+ connectString;
		
	    var loginDataLength=[]
	    loginDataLength.push32(loginData.length);
	    TERM.oob_socket.send(loginDataLength);
	    TERM.oob_socket.send_string(loginData); 
			
	}

	init_msg = function() {

		switch (term_state) {

		case 'connect':
		case 'OvmInit':

			var resultCode = TERM.socket.rQshift8();
			Util.Info("ResultCode from OvmInit : " + resultCode);

			switch (resultCode) {
			case 0:
				updateState('OvmWaitServerConnect',
						"Waiting for backend server connection");
				connect_oob();

				break;
			case 1:
				return fail("SessionId is invalid or timed out, reload the page with a fresh sessionid");
				break;
			case 3:
				// take the other chars
				var loggedInUserLength = TERM.socket.rQshift16();
				Util.Info("user length code " + loggedInUserLength);
				var loggedInUser = TERM.socket.rQshiftStr(loggedInUserLength);
				Util.Info("User " + loggedInUser);
				var confirm = window
						.confirm("Do you want to take over the connection from user "
								+ loggedInUser);
				Util.Info("Confirm window returned " + confirm);
				if (confirm == true) {
					TERM.socket.send([ 1 ]);
					updateState('OvmWaitServerConnect',
							"Waiting for backend server connection");
					connect_oob();
					
				} else {
					TERM.socket.send([ 0 ]);
					return fail("Connection in use!");
				}
				break;
			case 4:
    			return fail("VM has no assigned ServerPool or has no compatible servers to start on");
    			break;
			default:
				return fail("Unknown failure initing Ovm Connection, resultcode : "
						+ resultCode);
			}
			break;
		case 'OvmWaitServerConnect':

			var resultCode = TERM.socket.rQshift8();
			if (resultCode == 0) {
				updateState('normal', "Connected (secure) to : " + vmName);

			} else {
				updateState('OvmWaitServerConnect',
						"Waiting for backend server connection");
			}
			break;

		}

	};

	normal_msg = function() {
		if (viewer != undefined) {
			viewer.readBytes();
		} else {
			Util.Warn("viewer not initialized, not processing message");
		}
	};

	handle_message = function() {
		// Util.Debug(">> handle_message TERM.socket.rQlen(): "
		// + TERM.socket.rQlen());
		// Util.Debug("TERM.socket.rQslice(0,20): " + TERM.socket.rQslice(0, 20)
		// + " (" + TERM.socket.rQlen() + ")");
		if (TERM.socket.rQlen() === 0) {
			Util.Warn("handle_message called on empty receive queue");
			return;
		}

		switch (term_state) {
		case 'disconnected':
		case 'failed':
			Util.Error("Got data while disconnected");
			break;
		case 'normal':
			normal_msg();

			break;
		default:
			init_msg();
			break;
		}

	};
	


	handle_oob_message = function() {
	    //Util.Debug(">> handle_oob_message ws_oob.rQlen(): " + ws_oob.rQlen());
	    //Util.Debug("ws_oob.rQslice(0,20): " + ws_oob.rQslice(0,20) + " (" + ws_oob.rQlen() + ")");
		Util.Info(">> Handling Oob message");
	    if (TERM.oob_socket.rQlen() === 0) {
	        Util.Warn("handle_oob_message called on empty receive queue");
	        return;
	    }
	    
	    var msg_type = TERM.oob_socket.rQshift8();
	    Util.Info("Oob message type : " + msg_type + " received");
	    //Cases 0-3 : Backend trying to connect
	    //Case 96-99: Backend disconnected
	    switch (msg_type) {
	    case 0:
	    	// Back end server connected
	    	// We need to check here from where we come.
	    	Util.Info("Succesfull connection to backend ovs vm");
	    	
	    	updateState('normal', "Connected (secure) to : " + vmName);
	    	
	    	break;
	    case 1:
	    	onUpdateState('OvmTryingConnect', 'OvmWaitServerConnect','Server waiting to connect');
	    	break;
	    case 2: 
	    	onUpdateState('OvmTryingConnect', 'OvmWaitServerConnect','Server waiting to connect error 2');
	    	break;
	    case 30:
	    	updateState('disconnect', "VM deleted, disconnecting");
	    	break;
	    case 96:
	    	updateState('ovs_disconnect',"VM backend disconnected, VM is migrating to new server, connection will be restored when migrate is finished");
	    	break;
	    case 97:
	    	updateState('ovs_disconnect',"VM backend disconnected, VM has suspend, Retrying connection as soon as VM resumes again");
	    	break;
	    case 98:
	    	updateState('ovs_disconnect',"VM backend disconnected, VM has stopped, Retrying connection as soon as VM starts again");
	    	break;
	    case 99:
	    	updateState('ovs_disconnect',"Lost Vnc Server connection on backend, Trying to restore connection");
	    	break;
	    default:
	    	Util.Warn("Unrecognized Message type from Oob websocket : " + msg_type);
	    	break;
	    }
	    
	    Util.Info("<< Handling Oob message");
	}


	fail = function(msg) {
		updateState('failed', msg);
		return false;
	};
};